#include "ScoreMatrix.h"
#include "sparselib.hh"

using namespace ::sparselib_load ;
using           ::sparselib::index_type ;
using ::sparselib_fun::maxval ;
using ::sparselib_fun::minval ;
using ::sparselib_fun::absval ;

using namespace std;

#define DUMMY -1

class EPADLocal;

class EPADLocalSEQUENCE
{
public:
	EPADLocalSEQUENCE(){m_pModel=NULL; obs_label=NULL; predicted_label=NULL;};
	EPADLocalSEQUENCE(int len, EPADLocal* pModel);
	~EPADLocalSEQUENCE();

	void Init(int length_seq, EPADLocal* pModel);

	EPADLocal* m_pModel;

	int length_seq;

	int* obs_label;
	Score *_features;
	Score **obs_feature;

	Score Partition;
	int *predicted_label;
	ScoreMatrix *forward;
	ScoreMatrix *backward;

	ScoreMatrix* arrVi;
	ScoreMatrix gates;
	Score Gate(Score sum){ return (Score)1.0/(1.0+exp(-(double)sum)); }
	Score GetGateOutput(int gate, Score* features);
	void ComputeGates();
	void ComputeVi4Sampling();

	void ComputeForward();
	void ComputeBackward();
	void CalcPartition();

	void ComputePartition();
	Score ComputeScore(int leftState, int currState, int pos);

	void makeFeatures();
	Score* getFeatures(int pos);
};	

typedef map<string, int> mapStr2Int;
class EPADLocal
{
public:
	EPADLocal(){
		weights=NULL;
		labelMapI2S.clear();
		labelMapS2I.clear();
	};
	~EPADLocal(){
		delete weights;
	}

	int num_states;
	int num_labels;
	int num_data;
	int num_tst;
	int num_gates;
	int num_gates2;
	int dim_one_pos;
	int dim_features;
	int window_size;
	int num_params;
	int totalPos;
	int totalCorrect;

	string model_file;

	double* weights;

	// get the transition prob, num_states( for DUMMY ) + num_states*num_states
	inline Score GetWeightT(int leftState, int currLabel){return weights[num_labels * (leftState +1)+ currLabel];}

	// get the Gate weight
	inline Score GetWeightG(int gate, int dim){return weights[iFeatureStart+gate*dim_features + dim];};

	CCoorMatrix<int> sparseMi1, sparseMi;
	CCoorMatrix<double> d_sparseMi, d_sparseVi;
	sparselib::index_const_pointer  I1;
	sparselib::index_const_pointer  J1;
	int const* A1;
	sparselib::index_const_pointer  I;
	sparselib::index_const_pointer  J;
	int const* A;
	sparselib::index_const_pointer  d_I;
	sparselib::index_const_pointer  d_J;
	double const* d_A;
	CRowMatrix<double>  crow;
	sparselib::index_const_pointer  rowR;
	sparselib::index_const_pointer  rowJ;
	double const* rowA;
	CColMatrix<double>  ccol;
	sparselib::index_const_pointer  colI;
	sparselib::index_const_pointer  colC;
	double const* colA;
	sparselib::index_const_pointer  IV;
	sparselib::index_const_pointer  JV;
	int const* AV;

	int nnz_param;

	int iGateStart, iGate2Start, iFeatureStart;

	ScoreMatrix Mi;
	ScoreMatrix arrVi, memory;
	vector<string> labelMapI2S; 
	mapStr2Int labelMapS2I;
	int seq_length;
	EPADLocalSEQUENCE seq4sampling;

	void printVi(int pos);

public:
	void ComputeVi4Sampling();

	vector<string> temp_labels;

	int *m, *m0, *mm, *mmap;
	int *labels, *states;
	double *scale;

	void Initialize(string strModelFile, string strLabelFile, string strDataFile);
	double Calculate1orderEnergy(vector<int >& labels, int start, int end);
	void print_NormalizeViMatrixO1();
	void print_NormalizeTransitionMatrixO1();

protected:
	double* referenceState;
	ScoreMatrix residue_energy;
	double CaculateResidueEnergy(int idx, int label);
	void LoadSamplerData(string strDataFile);
	void addLabel(int l, string s){
		if ((int)labelMapI2S.size()>l) {
			labelMapI2S[l]=s;
			labelMapS2I.insert(pair<string, int>(s, l));
		}
	};
	int getLabelIndex(string aname){
		mapStr2Int::iterator it = labelMapS2I.find(aname);
		if (it == labelMapS2I.end())
			return -1;
		else 
			return it->second;
	};
};
