#include "ScoringFunction.h"

#include <vector>
#include <values.h>
#include <iostream>
#include <fstream>

#include "rmsdp.h"

using namespace std;

#define SQUARE(x) ((x)*(x))
#define PI 3.14159265358979

/************ScoringFunction is the base class for all the scoring methods *****************/
ScoringFunction::ScoringFunction(){
	startRes=1;
}

ScoringFunction::~ScoringFunction(){
}

int ScoringFunction::Init(string , string , string ){
	return 0;
}

double ScoringFunction::evaluate(vector<RESIDUE >& structure, int start, int end){
	return 0;
}
double ScoringFunction::evaluate(vector<RESIDUE >& structure){
	return 0;
}

//calculate the radius of a given protein
double ScoringFunction::CalculateRadius(vector<RESIDUE >& structure)
{
	double d, dist=0;
	for(int i=0; i<(int)structure.size();i++) {
		ATOM CAi = structure[i].getAtom(0);
		for(int j=0; j<i; j++) {
			ATOM CAj = structure[j].getAtom(0);
			d=(CAi.x-CAj.x)*(CAi.x-CAj.x)+(CAi.y-CAj.y)*(CAi.y-CAj.y)+(CAi.z-CAj.z)*(CAi.z-CAj.z);
			dist+=d;
		}
	}
	dist/=(SQUARE(structure.size())-structure.size());
	dist=sqrt(dist);
	return dist;
}

/****************** ScoringByEPAD *******************************/
ScoringByEPADLocal::ScoringByEPADLocal(){
	mName="EPADLocal";
}

ScoringByEPADLocal::~ScoringByEPADLocal(){
	for(int i = 0; i < libLen; i++) {
		for(int j = 0; j < fragLen; j++) 
			delete  fragments[i][j];
		delete fragments[i];
	}
	delete fragments;
}

// This function is necessary for initialization, 'coz DOPE need to read the dope.par for computation 
int ScoringByEPADLocal::Init(string modelFile, string feature_file, string labelFile)
{
	readLib();
	model.Initialize(modelFile, labelFile, feature_file);
	return 1;
}

void ScoringByEPADLocal::readLib()
{
	fragLen=5;
	libLen=100;
	fragLibFile="./config/lib_100_z_5.txt";

	// should add fragment in class def
	fragments = new double**[libLen];
	for(int i = 0; i < libLen; i++) {
		fragments[i] = new double*[fragLen];
		for(int j = 0; j < fragLen; j++)
			fragments[i][j] = new double[3];
	}
	char buf[4096];
	ifstream libin(fragLibFile.c_str());
	//debug
	if(!libin.is_open()) {
		cerr << " Can not open the fragment library!" << endl << endl;
		exit(1);
	}

	cerr << "============= start to read lib======" << endl;
	//skip header
	for(int i = 0; i < 20; i++)
		libin.getline(buf,4096);
	//read fragments
	for(int i = 0; i < libLen; i++){
		for(int j = 0; j < fragLen; j++) {
			while(libin.getline(buf,4096) && (buf[0]=='*' || buf[3] =='-'));
			sscanf(buf,"%lf%lf%lf",&fragments[i][j][0],&fragments[i][j][1],&fragments[i][j][2]);
		}
	}
	libin.close();
}
// Given a structure, assign labels to this structure
void ScoringByEPADLocal::ReLabel(int start, int end, BACKBONE& protein, vector<int >& labels)
{
	cerr << "\nReLabel" << endl;
	double **frag = new double*[fragLen];
	for(int j = 0; j < fragLen; j++)
		frag[j] = new double[3];
	for (int i = start; i <= end; i++) {
		// Assign Label
		for(int j = 0; j < fragLen; j++) {
  			// build the frag of length, whose center is the current target residue
			ATOM atm=protein[i-2+j].getAtom(0); // CA
  			frag[j][0] = atm.x;
  			frag[j][1] = atm.y;
  			frag[j][2] = atm.z;
		}
		double min_rmsd = 10000.;
		int index = 0;
		//find the most similar fragments
		for(int k = 0; k < libLen; k++) {
	      		double rmsd = 10000.;
			RMSD(fragments[k], frag, fragLen, &rmsd);
			if(rmsd < min_rmsd) {
				index = k;
				min_rmsd = rmsd;
			}
		}
		labels[i] = index;
	} 
	for(int j = 0; j < fragLen; j++)
		delete frag[j];
	delete frag;
}

double ScoringByEPADLocal::evaluate(vector<RESIDUE >& structure)
{
	return evaluate(structure, 0, -1);
}

double ScoringByEPADLocal::evaluate(vector<RESIDUE >& structure, int start, int end)
{
	int len=structure.size();
	if (end<start)
		end = structure.size();
 
	start=2;
	end=end-3;
	vector<int > labels(end+3);
	labels[0]=0;
	labels[1]=0;
	labels[end+1]=5;
	labels[end+2]=5;
	ReLabel(start, end, structure, labels);
	double energy= model.Calculate1orderEnergy(labels, start+1, end-1);
	if (energy>1e20)
		energy=0;
	cerr << " EPADLocal=" << energy << endl;
	return energy*0.0015*50;
}

/****************** ScoringByEPAD *******************************/
ScoringByEPAD::ScoringByEPAD(){
	mName="EPAD";
	startRes=-1;
	endRes=-1;
	predicted=false;
	indicator=0; // CA atoms
	mfunc=FUNC_CONTACT_ENERGY_TB; // For Rosetta
}

ScoringByEPAD::~ScoringByEPAD(){
}

// This function is necessary for initialization, 'coz DOPE need to read the dope.par for computation 
int ScoringByEPAD::Init(string config_file, string feature_file, string tgt_file)
{
	model.Initialize(config_file, feature_file);
	model.reset(startRes, endRes, residue_flag);
	NEFF=model.NEFF;
	model_typ=model.model_typ;
	if (startRes>-1 && endRes>-1 || mfunc==FUNC_CONTACT_DISTRIBUTION || mfunc==FUNC_CONTACT_PREDICTED_ENERGIES){
		model.testData->PredictDistances(mfunc);
		predicted=true;
	}
	return 1;
}

double ScoringByEPAD::evaluate(vector<RESIDUE >& structure)
{
	return evaluate(structure, 0, -1);
}

double ScoringByEPAD::evaluate(vector<RESIDUE >& structure, int start, int end)
{
	int len=structure.size();
	if (end<start)
		end = structure.size();
	endRes=end;

	model.reset(startRes, endRes, residue_flag);
	if (!predicted) {
		model.residue_flag.clear();
		for (int i=0; i<(int)structure.size(); i++){
			//cerr << i <<":" << structure[i].name<< endl;
			if (structure[i].name=="")
				model.residue_flag.push_back(0);
			else
				model.residue_flag.push_back(1);
		}
		model.testData->PredictDistances(mfunc);
		predicted=true;
	}

	int contact_dist_min;
	switch (mfunc)
	{
	case FUNC_CONTACT_ENERGY_FM:
		contact_dist_min=6;
		break;
	case FUNC_CONTACT_ENERGY_TB:
		contact_dist_min=7; // temporarily change to 6 (2012/02/29)
		if (len<85)
			contact_dist_min=6;
		break;
	case FUNC_CONTACT_PAIRWISE_ENERGY:
	case FUNC_CONTACT_DISTRIBUTION_ENERGY:
		contact_dist_min=6;
		break;
	default:
		if ((int)structure.size()>65)
			contact_dist_min=12;
		else
			contact_dist_min=9;
		break;
	}

	cout.setf(ios::fixed,ios::floatfield);
	cout.precision(3);

	double rg = CalculateRadius(structure);
	cerr << mfunc << "::::::::::::::::: " << contact_dist_min << " start="<< start << ";  end="<<end << " structure length=" << structure.size()<< endl;
	double energy = 0.0;
	for(int i=0; i<len-1; i++) {
		if (structure[i].seq==-1) continue;
		ATOM atom_i;
		if (indicator==0)
			atom_i = structure[i].getAtom(0);
		else if (structure[i].name.compare("GLY")==0)
			atom_i = structure[i].getAtom(0);
		else
			atom_i = structure[i].getAtom(1);
		
		double residue_energy = 0.0;
		for(int j=i+contact_dist_min; j<len; j++) {
			if (structure[i].name_idx<0 || structure[j].name_idx<0)
				continue;
			int cd = abs(i-j); //chainDist
			ATOM atom_j;
			if (indicator==0)
				atom_j=structure[j].getAtom(0);
			else if (structure[j].name.compare("GLY")==0)
				atom_j=structure[j].getAtom(0);
			else
				atom_j=structure[j].getAtom(1);
			
			double dist= B_B_Q::dist(atom_i, atom_j);

			int ri=structure[i].name_idx, rj=structure[j].name_idx;
			double e_pair = model.CalculatePairEnergyS12(structure[i].seq, structure[j].seq, dist, rg, mfunc, ri, rj);
			residue_energy += e_pair;
		}
		energy +=residue_energy;
		//cerr << "energy = " << energy << endl;
	}
	if (mfunc==FUNC_CONTACT_PAIRWISE_ENERGY || model_typ==0) // Folding only need Ca-Ca interactions
		return energy*0.0015*50;

	double energyB = 0.0;
	double energyA = 0.0;
	energy=energy*0.0015*50*0.2;
//*
	for(int i=0; i<len-1; i++) {
		if (structure[i].seq==-1 || model.cond_weight==0 || len<45 || model_typ==3) continue;
		RESIDUE resi=structure[i];
		double residue_energyA = 0.0;
		double residue_energyB = 0.0;
		for(int j=i+contact_dist_min; j<(int)structure.size(); j++) {
			RESIDUE resj=structure[j];
			for(int ai=0; ai<(int)resi.arrAtoms.size();ai++){
				if (ai>4)
					continue;
				ATOM atom_i = resi.arrAtoms[ai];
				int atom1=atom_i.name_idx;
				for(int aj=0; aj<(int)resj.arrAtoms.size();aj++){
					if (aj>4)
						continue;
					ATOM atom_j = resj.arrAtoms[aj];
					int atom2=atom_j.name_idx;
					if (!(atom1==2&&atom2==0 || atom1==2&&atom2==1 || (atom1==0||atom1==1||atom1==2||atom1==3)&&atom2==2))
						continue;
					double dist = B_B_Q::dist(atom_i, atom_j);
					int ri=structure[i].name_idx, rj=structure[j].name_idx;
					double e_pair=model.CalculateConditionalPairEnergyS12(structure[i].seq,structure[j].seq,dist,rg,mfunc,ri,rj,atom1,atom2);
					//cerr << "distance(" << atom1 << "," << atom2 << ")= " << dist << "; e_pair=" << e_pair << endl;
					residue_energyB += e_pair;
				}
			}
		}
		energyB +=residue_energyB;
		//energyA +=residue_energyA;
	}
	energyA=0;
	energyB=energyB*0.0015*model.cond_weight;
	cerr << " DPN energy=" << energy << "; energyB=" << energyB << endl;
// */	
	//model.printConfusionMatrix(rg);
	return (energy+energyA+energyB);
}
