
#ifndef  _CONSTRAINTS_H
#define _CONSTRAINTS_H

#include <vector>
#include "ScoreMatrix.h"

using namespace std;

struct Coefficient
{
	int idx;
	double val;
	Coefficient(int idx, double val){ this->idx=idx; this->val=val;}
};

class Constraint 
{
public:
	vector<Coefficient > co;
	void setDist(double val, int i, int j, int seq_len){
		if (i>j){int tmp=i; i=j; j=tmp;}
		int idx=i*seq_len-i*(3+i)/2+j; // idx(i,j)=(j-i)+sum(l-t-1), 0<=t<=i-1 ==> idx=il-(3+i)*i/2)+j
		Coefficient c(idx,val);
		co.push_back(c);
	}
	void setBeta(double val, int r, int s, int max_beta, int beta_start) {
		if (r>s){int tmp=r; r=s; s=tmp;}
		int idx=beta_start+r*max_beta-r*(3+r)/2+s;
		Coefficient c(idx,val);
		co.push_back(c);
	}
	void setConstB(int idx, double val){
		Coefficient c(idx,val);
		co.push_back(c);
	}
};

struct AlphaStruct
{
	int start, end;
};

struct BetaStruct
{
	int start, end;
};

class Constraints
{
public:
	Constraints();
	~Constraints(){};

	void getSecondaryStructure(int seq_len, const int* ss);
	void generateConstraints(int seq_len, const int* ss, string name);
	void print(int seq_len, int assistVarNum, int num_alpha, int num_beta);
	int getSSIdx(int residue); // get the secondary structure index of a residue

	vector<Constraint > constraintMat;

	int beta_start, alpha_start, totalVar;

private:
	string p_name;
	int seq_len;
	vector<BetaStruct > betas;
	vector<AlphaStruct > alphas;
	vector<int > ss_idx; // for each residue, we assign it i if it's on the i-th strand, and 100+i if it's on the i-th helix, -1 for loop
	int beta_idx, alpha_idx; // for alpha, 99=100-1 denotes -1
};

#endif
