#ifndef _BBQ_H
#define _BBQ_H

#include <fstream>
#include <vector>
#include <string>
#include <iostream>
#include <cmath>
#include <map>
#include <dirent.h>
#include <errno.h>

using namespace std;
#define HELIX	0
#define STRAND	1
#define COIL	2
#define TURN	2

class ATOM
{
public:
	string name;
	int name_idx; // CA=0; CB=1; O=2; N=3; C=4; HN=5; SG=6
	double x,y,z;
	int res_idx;

	ATOM(){x=y=z=0.0;name="";res_idx=-1;name_idx=-1;};
	ATOM(string aname, int i){
		this->name=aname;x=y=z=0.0;res_idx=i; name_idx=-1;
		string arrAtomName[7]={"CA", "CB", "O", "N", "C", "HN", "SG"};
		for (int n=0; n<7; n++){
			if (arrAtomName[n]==aname){
				name_idx=n;
				break;
			}
		}
	};
};

// map of atoms' name [string -> int]
typedef map<string, int > mapStr2ATOM;

class RESIDUE
{
public:
	RESIDUE(){ phi=0; psi=0;seq=-1; name=""; name_idx=-1;
		for (int i=0; i<10; i++)
			atomMap[i]=-1;
	}
	mapStr2ATOM atom_map;
	int freq;
	string name;
	int name_idx;
	double phi, psi;
	int ramaBasin;
	int seq;

	// return secondary structure
	int getSecondaryStructure() {
		if ((-180 < phi) && (phi < -20) && (-90 < psi) && (psi < -10)) return HELIX;
		if ((-180 < phi) && (phi < -20) && (((20 < psi) && (psi < 180)) || ((-170 < psi) && (psi < -180)))) return STRAND;
		return COIL;
	}

	void addAtom(ATOM atom){
		if (getAtomIndex(atom.name)==-1) {
			// cerr << seq << "-ATOM::addAtom(" << atom.name << ") " << endl;
			// atom not found, insert it to the map
			atom.res_idx = seq;
			int n = arrAtoms.size();
			arrAtoms.push_back(atom);
			atom_map.insert(pair<string, int>(atom.name, n));
			string arrAtomName[7]={"CA", "CB", "O", "N", "C", "HN", "SG"};
			for (int i=0; i<7; i++){
				if (arrAtomName[i]==atom.name){
					atomMap[i]=n;
					break;
				}
			}
		}
	};
	
	int getAtomIndex(string aname){
		//cerr << "-" << aname << seq;
		mapStr2ATOM::iterator atomit;
		atomit = atom_map.find(aname);
		if (atomit == atom_map.end()) {
			//cerr << "N" << endl;
			return -1;
		} else {
			//cerr << "f" << endl;
			return atomit->second;
		}
	};
	int getAtomIndex(int at){ return atomMap[at];};

	void deleteAtom(string aname){
		int idx=-1;
		string arrAtomName[7]={"CA", "CB", "O", "N", "C", "HN", "SG"};
		for (int i=0; i<7; i++){
			if (arrAtomName[i]==aname){
				idx=atomMap[i];
				atomMap[i]=-1;
				break;
			}
		}
		for (int i=0; i<10; i++) {
			if (atomMap[i]>idx)
				atomMap[i]--;
		}
		mapStr2ATOM::iterator atomit;
		atomit = atom_map.find(aname);
		if (atomit != atom_map.end()) { // atom found, delete it from the map
		  arrAtoms[atomit->second].name="";
		  arrAtoms.erase(arrAtoms.begin()+atomit->second);
		  atom_map.erase(atomit);
		}
		for(mapStr2ATOM::iterator at1=atom_map.begin();at1!=atom_map.end();at1++)
			if (at1->second>idx)
				at1->second--;
	};

	ATOM getAtom(string aname){
		//cerr << "." << arrAtoms.size();
		int i = getAtomIndex(aname);
		//cerr << "*" << i;
		if (i!=-1)
			return arrAtoms[i];
		else
			return ATOM();
	};
	ATOM getAtom(int at){
		int i = atomMap[at];
		if (i!=-1)
			return arrAtoms[i];
		else
			return ATOM();
	};
	vector<ATOM> arrAtoms;
	int atomMap[10];
};

//
// Global variables, KD-tree
//
class B_B_Q{
public:
	static double dist(ATOM a, ATOM b) {
	    ATOM c; 
	    c.x = a.x-b.x;
	    c.y = a.y-b.y;
	    c.z = a.z-b.z;
	    return sqrt( c.x*c.x + c.y*c.y + c.z*c.z);
	};

	static int lookup(vector<string>& files, const char* dir, const char *arg) {
		DIR *dirp;
		struct dirent *dp;
		if ((dirp = opendir(dir)) == NULL) {
			perror("couldn't open directory");
			return 0;
		}
		cerr << "Looking up " << arg << " files under the directory " << dir << endl;
		do {
			errno = 0;
			if ((dp = readdir(dirp)) != NULL) {
				char* p =  strstr(dp->d_name, arg);
				if (p!= 0) {	
					files.push_back(dp->d_name);
					continue;
				}
			}
		} while (dp != NULL);
	
		if (errno != 0)
			cerr <<"error reading directory"<<endl;
		(void) closedir(dirp);
		if (files.size()>0)
			return 1;
		else
			return 0;
	}
};

#endif
